/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.appManager;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.util.UUID;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.AppDescription;
import eu.DECIDEh2020.appManager.models.Microservice;
import org.everit.json.schema.Schema;
import org.everit.json.schema.ValidationException;
import org.everit.json.schema.loader.SchemaLoader;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class AppDescriptionFactory {

	private static Logger log = LoggerFactory.getLogger(AppDescriptionFactory.class);

	private static Schema schema;

	static {
		try (InputStream inputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream("application_description.schema.json")) {
			JSONObject rawSchema = new JSONObject(new JSONTokener(inputStream));
			schema = SchemaLoader.load(rawSchema);
		} catch (IOException e) {
			log.error("reading schema", e);
		}

		/*
		JsonSchemaFactory factory = JsonSchemaFactory.byDefault();
		InputStream is = Thread.currentThread().getContextClassLoader().getResourceAsStream("application_description.schema.json");
		try {
			schema = factory.getJsonSchema(new ObjectMapper().readTree(is));
		} catch (ProcessingException | IOException e) {
			log.error("reading json schema", e);
		}
		*/
	}

	public static AppDescription createAppDescription(String name) {
		AppDescription appDescription = new AppDescription();
		appDescription.setId(UUID.randomUUID().toString());
		appDescription.setName(name);

		appDescription.setVersion("0.2.2-alpha");
		appDescription.setHighTechnologicalRisk(false);

		appDescription.getMicroservices().add(createMicroservice("Default"));

		return appDescription;
	}

	public static Microservice createMicroservice(String name) {
		Microservice microservice = new Microservice();
		microservice.setId(UUID.randomUUID().toString());
		microservice.setName(name);

		microservice.getEndpoints().add("http://mircroservice1.com/service");

		return microservice;
	}

	public static AppDescription loadAppDescription(Path path) {
		try {
			return new ObjectMapper().readValue(path.toFile(), AppDescription.class);
		} catch (IOException e) {
			log.error("reading app description file", e);
			return null;
		}
	}

	public static AppDescription fromJson(String jsonString) throws JsonParseException, JsonMappingException, IOException {
		return new ObjectMapper().readValue(jsonString, AppDescription.class);
	}

	public static void validateAppDescription(AppDescription appDescription) throws DECIDEValidationException {
		try {
			schema.validate(new JSONObject(new ObjectMapper().writeValueAsString(appDescription)));
		} catch (JsonProcessingException e) {
			log.error("json to string", e);
			throw new DECIDEValidationException(e);
		} catch (ValidationException e) {
			log.error("validating json", e);
			throw new DECIDEValidationException(e);
		}
	}

	public static void validateAppDescription(JsonNode appDescription) throws DECIDEValidationException {
		try {
			validateAppDescription(new ObjectMapper().treeToValue(appDescription, AppDescription.class));
		} catch (JsonProcessingException e) {
			log.error("json to value", e);
			throw new DECIDEValidationException(e);
		}
	}

}
