/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.appManager;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;

import eu.DECIDEh2020.appManager.exceptions.AppManagerException;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.AppDescription;
import eu.DECIDEh2020.appManager.persistence.GitHandler;

import org.eclipse.jgit.api.errors.GitAPIException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Closeable;
import java.io.IOException;
import java.nio.file.Path;

public class AppManager implements Closeable {

    private Logger log = LoggerFactory.getLogger(getClass());

    private GitHandler gitHandler;

    private String descriptor = "DECIDE.json";

    private boolean strict = true;

    public static AppManager open(String gitRef, String username, String password, Path localPath) throws AppManagerException {
        return new AppManager(gitRef, username, password, localPath);
    }

    public static AppManager open(String gitRef, String token, Path localPath) throws AppManagerException {
        return new AppManager(gitRef, token, localPath);
    }

    public static AppManager open(Path localPath)throws AppManagerException {
        return new AppManager(localPath);
    }

    private AppManager() {
        gitHandler = new GitHandler(new GitHandler.Repository());
    }

    private AppManager(String gitRef, String token, Path localPath) throws AppManagerException {
        this();

        gitHandler.getRepository().setURI(gitRef);
        gitHandler.getRepository().setPath(localPath);

        gitHandler.setToken(token);

        try {
            gitHandler.openOrClone();
        } catch (IOException | GitAPIException e) {
            log.error("open or clone repo", e);
            throw new AppManagerException(e);
        }
    }

    private AppManager(String gitRef, String username, String password, Path localPath) throws AppManagerException {
        this();
        gitHandler.getRepository().setURI(gitRef);
        gitHandler.getRepository().setPath(localPath);

        gitHandler.setCredentials(username, password);
        try {
            gitHandler.openOrClone();
        } catch (IOException | GitAPIException e) {
            log.error("open or clone repo", e);
            throw new AppManagerException(e);
        }
    }
    
    private AppManager(Path localPath) throws AppManagerException {
        this();

        gitHandler.getRepository().setPath(localPath);

        try {
            gitHandler.openOrClone();
        } catch (IOException | GitAPIException e) {
            log.error("open or clone repo", e);
            throw new AppManagerException(e);
        }
    }

    public String setDescriptor(String descriptor) {
        String old = this.descriptor;
        this.descriptor = descriptor;
        return old;
    }

    public HistoryManager getHistoryManager() {
        return HistoryManager.create(gitHandler);
    }

    public AppDescription getAppDescription() throws DECIDEValidationException, IOException {
        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        AppDescription appDescription = new ObjectMapper().readValue(path.toFile(), AppDescription.class);
        if (strict) {
            AppDescriptionFactory.validateAppDescription(appDescription);
        }
        return appDescription;
    }

    public ObjectNode getAppDescriptionAsTree() throws DECIDEValidationException, IOException {
        return new ObjectMapper().valueToTree(getAppDescription());
    }

    public void write(AppDescription appDescription) throws AppManagerException, IOException {
        if (strict) {
            AppDescriptionFactory.validateAppDescription(appDescription);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), appDescription);

        try {
            gitHandler.addToRepo();
        } catch (GitAPIException e) {
            log.error("write app description", e);
            throw new AppManagerException(e);
        }
    }

    public void writeAndCommit(AppDescription appDescription, String commitMessage) throws AppManagerException, IOException {
        if (strict) {
            AppDescriptionFactory.validateAppDescription(appDescription);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), appDescription);

        try {
            gitHandler.addToRepo();
            gitHandler.commitToRepo(commitMessage);
        } catch (GitAPIException e) {
            log.error("write and commit app description", e);
            throw new AppManagerException(e);
        }
    }

    public void writeAndSync(AppDescription appDescription, String commitMessage) throws AppManagerException, IOException {
        if (strict) {
            AppDescriptionFactory.validateAppDescription(appDescription);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), appDescription);

        try {
            gitHandler.addToRepo();
            gitHandler.commitToRepo(commitMessage);
            gitHandler.pushToRepo();
        } catch (GitAPIException e) {
            log.error("synchronize app description", e);
            throw new AppManagerException(e);
        }
    }

    public void sync() throws AppManagerException {
        try {
            gitHandler.pullFromRepo();
        } catch (GitAPIException e) {
            log.error("synchronize app description", e);
            throw new AppManagerException(e);
        }
    }

    public void reset() {
        gitHandler.resetAll();
    }

    @Override
    public void close() throws IOException {
        try {
            gitHandler.addToRepo();
            gitHandler.commitToRepo("Closed by AppManager.");
            gitHandler.pushToRepo();
        } catch (GitAPIException e) {
            throw new IOException(e);
        } finally {
            gitHandler.close();
        }
    }

    public boolean isStrict() {
        return strict;
    }

    public void setStrict(boolean strict) {
        this.strict = strict;
    }

}
