/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.appManager;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import eu.DECIDEh2020.appManager.exceptions.AppManagerException;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.HistoryEntry;
import eu.DECIDEh2020.appManager.persistence.GitHandler;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;

public class HistoryManager {

    private Logger log = LoggerFactory.getLogger(getClass());

    private GitHandler gitHandler;

    private String descriptor = "OPTIMUSHistory.json";

    private boolean strict = true;

    static HistoryManager create(GitHandler gitHandler) {
        return new HistoryManager(gitHandler);
    }

    private HistoryManager(GitHandler gitHandler) {
        this.gitHandler = gitHandler;
    }

    public String getDescriptor() {
        return descriptor;
    }

    public void setDescriptor(String descriptor) {
        this.descriptor = descriptor;
    }

    public boolean isStrict() {
        return strict;
    }

    public void setStrict(boolean strict) {
        this.strict = strict;
    }

    public List<HistoryEntry> getHistory() throws DECIDEValidationException, IOException {
        return Arrays.asList(new ObjectMapper().treeToValue(getHistoryAsTree(), HistoryEntry[].class));
    }

    public List<HistoryEntry> getHistory(Path path) throws AppManagerException, IOException {
        return HistoryFactory.loadHistory(path);
    }

    public ArrayNode getHistoryAsTree() throws DECIDEValidationException, IOException {
        ObjectMapper mapper = new ObjectMapper();
        JsonNode history = mapper.createArrayNode();
        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        if (Files.exists(path)) {
            history = mapper.readTree(path.toFile());
            if (strict) {
                HistoryFactory.validateHistory(history);
            }
        } else {
            mapper.writerWithDefaultPrettyPrinter().writeValue(path.toFile(), history);
        }
        return (ArrayNode)history;
    }

    public void write(List<HistoryEntry> history) throws AppManagerException, IOException {
        if (strict) {
            HistoryFactory.validateHistory(history);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), history);

        try {
            gitHandler.addToRepo();
        } catch (GitAPIException e) {
            log.error("write app description", e);
            throw new AppManagerException(e);
        }
    }

    public void writeAndCommit(List<HistoryEntry> history, String commitMessage) throws AppManagerException, IOException {
        if (strict) {
            HistoryFactory.validateHistory(history);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), history);

        try {
            gitHandler.addToRepo();
            gitHandler.commitToRepo(commitMessage);
        } catch (GitAPIException e) {
            log.error("write and commit app description", e);
            throw new AppManagerException(e);
        }
    }

    public void writeAndSync(List<HistoryEntry> history, String commitMessage) throws AppManagerException, IOException {
        if (strict) {
            HistoryFactory.validateHistory(history);
        }

        Path path = gitHandler.getRepository().getPath().resolve(descriptor);
        new ObjectMapper().writerWithDefaultPrettyPrinter().writeValue(path.toFile(), history);

        try {
            gitHandler.addToRepo();
            gitHandler.commitToRepo(commitMessage);
            gitHandler.pushToRepo();
        } catch (GitAPIException e) {
            log.error("synchronize app description", e);
            throw new AppManagerException(e);
        }
    }

}
