/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.appManager.persistence;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Iterator;

import org.eclipse.jgit.api.*;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.api.errors.InvalidRemoteException;
import org.eclipse.jgit.api.errors.JGitInternalException;
import org.eclipse.jgit.merge.MergeStrategy;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.transport.*;
import org.eclipse.jgit.util.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class GitHandler {

    private final Logger log = LoggerFactory.getLogger(getClass());

	public static class Repository {
		Path path;

		private String uri, host, user, password, name, port, protocol;
		private boolean newRepo = true, localRepo = true;
		private String token;

		public Repository(String uri, Path path, String host, String user, String password, String port, String name, String protocol) {
			this.uri = uri;
			this.path = path;
			this.host = host;
			this.user = user;
			this.name = name;
			this.port = port;
			this.protocol = protocol;
			this.password = password;
		}

		public Repository() {

		}

		/**
		 * @return the URI
		 */
		public String getURI() {
			return uri;
		}

		/**
		 * @param uri
		 *            the URI to set
		 */
		public void setURI(String uri) {
			this.uri = uri;
		}

		/**
		 * @return the path
		 */
		public Path getPath() {
			return path;
		}

		/**
		 * @param path
		 *            the path to set
		 */
		public void setPath(Path path) {
			this.path = path;
		}

		/**
		 * @return the host
		 */
		public String getHost() {
			return host;
		}

		/**
		 * @param host
		 *            the host to set
		 */
		public void setHost(String host) {
			this.host = host;
		}

		/**
		 * @return the user
		 */
		public String getUser() {
			return user;
		}

		/**
		 * @param user
		 *            the user to set
		 */
		public void setUser(String user) {
			this.user = user;
		}

		/**
		 * @return the password
		 */
		public String getPassword() {
			return password;
		}

		/**
		 * @param password
		 *            the password to set
		 */
		public void setPassword(String password) {
			this.password = password;
		}

		/**
		 * @return the name
		 */
		public String getName() {
			return name;
		}

		/**
		 * @param name the name to set
		 */
		public void setName(String name) {
			this.name = name;
		}

		/**
		 * @return the port
		 */
		public String getPort() {
			return port;
		}

		/**
		 * @param port the port to set
		 */
		public void setPort(String port) {
			this.port = port;
		}

		/**
		 * @return the protocol
		 */
		public String getProtocol() {
			return protocol;
		}

		/**
		 * @param protocol the protocol to set
		 */
		public void setProtocol(String protocol) {
			this.protocol = protocol;
		}

		/**
		 * @return the newRepo
		 */
		public boolean isNewRepo() {
			return newRepo;
		}

		/**
		 * @param newRepo the newRepo to set
		 */
		public void setNewRepo(boolean newRepo) {
			this.newRepo = newRepo;
		}

		/**
		 * @return the localRepo
		 */
		public boolean isLocalRepo() {
			return localRepo;
		}

		/**
		 * @param localRepo the localRepo to set
		 */
		public void setLocalRepo(boolean localRepo) {
			this.localRepo = localRepo;
		}

		public String getToken() {
			return token;
		}

		public void setToken(String token) {
			this.token = token;
		}
	}
	
	private Repository repository;
	private Git git;
	private CredentialsProvider cp;
	
	public GitHandler(Repository repository) {
		this.repository = repository;
		if(repository.getName() != null && !repository.getName().isEmpty()
				&& repository.getPassword()!=null && !repository.getPassword().isEmpty()) {
			cp=new UsernamePasswordCredentialsProvider(repository.getName(), repository.getPassword());
		}
		
	}

	public void openOrClone() throws IOException, GitAPIException {
		FileRepositoryBuilder repositoryBuilder = new FileRepositoryBuilder();
		repositoryBuilder.addCeilingDirectory(repository.getPath().toFile());
		repositoryBuilder.findGitDir(repository.getPath().toFile());

		if( repositoryBuilder.getGitDir() == null ) {
			git = Git.cloneRepository()
					.setURI(repository.getURI())
					.setCredentialsProvider(cp)
					.setDirectory(repository.getPath().toFile())
					.call();
		} else {
			git = new Git(repositoryBuilder.build());
		}
	}


	public void cloneRepositoryTo(Path dir) throws GitAPIException {
		repository.setPath(dir);
		cloneRepo();
	}
	
	public void createRepositoryIn(Path dir) throws IllegalStateException, GitAPIException {
		git = Git.init().setDirectory(dir.toFile()).call();
		repository.setPath(git.getRepository().getDirectory().toPath());
	}
	
    public void cloneRepo() throws GitAPIException {
		git = Git.cloneRepository()
				.setURI(repository.getURI())
				.setCredentialsProvider(cp)
                .setDirectory(repository.getPath().toFile())
                .call();
    }

    public void addToRepo() throws GitAPIException {
        AddCommand add = git.add();
        add.addFilepattern(".").call();
    }

    public void commitToRepo(String message) throws JGitInternalException, GitAPIException {
        git.commit().setMessage(message).call();
    }

    public void pushToRepo() throws JGitInternalException, GitAPIException {
        PushCommand pc = git.push();
        pc.setCredentialsProvider(cp)
                .setForce(true)
                .setPushAll();
        try {
            Iterator<PushResult> it = pc.call().iterator();
            if (it.hasNext()) {
            	log.debug(it.next().toString());
            }
        } catch (InvalidRemoteException e) {
            log.error("push to repo", e);
        }
    }

    public void pullFromRepo() throws GitAPIException {
		try {
			PullResult result = git.pull().setCredentialsProvider(cp).setStrategy(MergeStrategy.RESOLVE).call();
			if (result.isSuccessful()) {
				return;
            }
		} catch (GitAPIException e) {
            log.error("standard pull", e);
		}
        resetMerging();
        git.pull().setStrategy(MergeStrategy.THEIRS).setCredentialsProvider(cp).call();
    }

    public void setCredentials(String name,String password) {
		repository.setUser(name);
		repository.setPassword(password);
		prepareAuthentication();
    }
    
	public void setToken(String token) {
		repository.setToken(token);
		prepareAuthentication();
	}

	/**
	 * @return the repository
	 */
	public Repository getRepository() {
		return repository;
	}

	/**
	 * @param repository the repository to set
	 */
	public void setRepository(Repository repository) {
		this.repository = repository;
	}

	private void prepareAuthentication() {
		try {
			URL url = new URL(repository.getURI());
			String protocol = url.getProtocol();
			String host = url.getHost();
			String token = repository.getToken();

			if (protocol.equalsIgnoreCase("https") && token != null && !token.isEmpty()) {
				if (host.contains("gitlab") || host.contains("tecnalia")) {
					String newUrl = protocol + "://" + "gitlab-ci-token:" + URLEncoder.encode(token, "UTF-8") + "@" + url.getAuthority() + url.getFile();
					if (url.getRef() != null) {
						newUrl += "#" + url.getRef();
					}
					repository.setURI(newUrl);
					cp = new UsernamePasswordCredentialsProvider("username", token);
				} else if (host.contains("github")) {
					cp = new UsernamePasswordCredentialsProvider(token, "");
				} else {
					cp = new UsernamePasswordCredentialsProvider(repository.getUser(), repository.getPassword());
				}
			} else {
				cp = new UsernamePasswordCredentialsProvider(repository.getUser(), repository.getPassword());
			}

		} catch (MalformedURLException e) {
			log.error("repo url parsing", e);
		} catch (UnsupportedEncodingException e) {
			log.error("token encoding", e);
		}
	}

	private void resetMerging() throws GitAPIException {
		try {
			git.getRepository().writeMergeCommitMsg( null );
			git.getRepository().writeMergeHeads( null );
		} catch (IOException e) {
			log.error("abort merging", e);
		}
		git.reset().setMode( ResetCommand.ResetType.HARD ).call();
	}

	public void resetAll() {
		git.getRepository().close();
		close();
		try {
			FileUtils.delete(repository.getPath().toFile(), FileUtils.RECURSIVE | FileUtils.IGNORE_ERRORS);
			log.debug("directory deleted");
			log.debug("exists {}", Files.exists(repository.getPath()));
			cloneRepo();
		} catch (GitAPIException | IOException e) {
			log.error("reset cloned repo", e);
		}
	}

	public void close() {
		if (git != null) {
			git.close();
		}
	}

}
